"""Defines the Pydantic model for the URDF to MJCF conversion."""

import enum
from typing import Literal

from pydantic import BaseModel

Angle = Literal["radian", "degree"]


class CollisionParams(BaseModel):
    condim: int = 3
    contype: int = 0
    conaffinity: int = 1
    priority: int = 1
    solimp: list[float] = [0.99, 0.999, 0.00001]
    solref: list[float] = [0.005, 1.0]
    friction: list[float] = [1.0, 0.01, 0.01]


class JointMetadata(BaseModel):
    actuator_type: str | None = None
    id: int
    nn_id: int
    kp: float
    kd: float
    soft_torque_limit: float
    min_angle_deg: float | None = None
    max_angle_deg: float | None = None

    @classmethod
    def from_dict(cls, data: dict) -> "JointMetadata":
        """Create JointParam from a plain dictionary."""
        return cls(**data)

    class Config:
        extra = "forbid"


class ActuatorMetadata(BaseModel):
    actuator_type: str
    sysid: str = ""
    max_torque: float | None = None
    max_velocity: float | None = None
    armature: float | None = None
    damping: float | None = None
    frictionloss: float | None = None
    vin: float | None = None
    kt: float | None = None
    R: float | None = None
    max_pwm: float | None = None
    error_gain: float | None = None

    @classmethod
    def from_dict(cls, data: dict) -> "ActuatorMetadata":
        """Create ActuatorParam from a plain dictionary."""
        return cls.model_validate(data)


class ImuSensor(BaseModel):
    body_name: str
    pos: list[float] | None = None
    rpy: list[float] | None = None
    acc_noise: float | None = None
    gyro_noise: float | None = None
    mag_noise: float | None = None


class CameraSensor(BaseModel):
    name: str
    mode: str
    pos: list[float] | None = None
    rpy: list[float] | None = None
    fovy: float = 45.0


class ForceSensor(BaseModel):
    """Represents a force sensor attached to a site."""

    body_name: str
    site_name: str
    name: str | None = None
    noise: float | None = None


class ExplicitFloorContacts(BaseModel):
    """Add explicit floor contacts."""

    contact_links: list[str]
    class_name: str = "collision"


class WeldConstraint(BaseModel):
    """Represents a weld constraint between two bodies.

    Useful for locking bodies in the air (suspending).
    By default, body2 is set to "world" to create a fixed constraint in global space.
    Default parameters approximately lock the body rigidly in place -- stiffer than Mujoco defaults.
    """

    body1: str
    body2: str = "world"
    solimp: list[float] = [0.95, 0.99, 0.005, 0.5, 2]  # Very tight width, aggressive response.
    solref: list[float] = [0.005, 1.0]  # critical damping 1.0, very quick time constant to return to static


class CollisionType(enum.Enum):
    BOX = enum.auto()
    PARALLEL_CAPSULES = enum.auto()
    CORNER_SPHERES = enum.auto()
    SINGLE_SPHERE = enum.auto()


class CollisionGeometry(BaseModel):
    name: str
    collision_type: CollisionType
    sphere_radius: float = 0.01
    axis_order: tuple[int, int, int] = (0, 1, 2)
    flip_axis: bool = False
    offset_x: float = 0.0
    offset_y: float = 0.0
    offset_z: float = 0.0


class ConversionMetadata(BaseModel):
    freejoint: bool = True
    collision_params: CollisionParams = CollisionParams()
    joint_name_to_metadata: dict[str, JointMetadata] | None = None
    actuator_type_to_metadata: dict[str, ActuatorMetadata] | None = None
    imus: list[ImuSensor] = []
    cameras: list[CameraSensor] = [
        CameraSensor(
            name="front_camera",
            mode="track",
            pos=[0, 2.0, 0.5],
            rpy=[90.0, 0.0, 180.0],
            fovy=90,
        ),
        CameraSensor(
            name="side_camera",
            mode="track",
            pos=[-2.0, 0.0, 0.5],
            rpy=[90.0, 0.0, 270.0],
            fovy=90,
        ),
    ]
    force_sensors: list[ForceSensor] = []
    collision_geometries: list[CollisionGeometry] | None = None
    explicit_contacts: ExplicitFloorContacts | None = None
    weld_constraints: list[WeldConstraint] = []
    remove_redundancies: bool = True
    floating_base: bool = True
    maxhullvert: int | None = None
    angle: Angle = "radian"
    floor_name: str = "floor"
    add_floor: bool = False
    backlash: float | None = None
    backlash_damping: float = 0.01
    height_offset: float = 0.0
    visualize_collision_meshes: bool = True

    class Config:
        extra = "forbid"



"""Defines a post-processing function that adds backlash joints to hinge joints."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from urdf2mjcf.utils import save_xml

logger = logging.getLogger(__name__)


def add_backlash_default(root: ET.Element, backlash: float, backlash_damping: float) -> None:
    """Add a default class for backlash joints.

    Args:
        root: The root element of the MJCF file.
        backlash: The amount of backlash to add.
        backlash_damping: The damping value for backlash joints.
    """
    # Check if default element exists
    default = root.find("default")
    if default is None:
        default = ET.SubElement(root, "default")

    # Check if backlash class already exists
    backlash_default = default.find(".//default[@class='backlash']")
    if backlash_default is None:
        backlash_default = ET.SubElement(default, "default", attrib={"class": "backlash"})

        # Add joint properties for backlash
        # These values are based on the reference example
        ET.SubElement(
            backlash_default,
            "joint",
            attrib={
                "damping": f"{backlash_damping:.6g}",
                "frictionloss": "0",
                "armature": "0.01",
                "limited": "true",
                "range": f"{-backlash:.6g} {backlash:.6g}",
            },
        )

        logger.info("Added backlash default class")


def find_parent_body(joint: ET.Element, root: ET.Element) -> ET.Element | None:
    """Find the parent body element of a joint.

    Args:
        joint: The joint element.
        root: The root element of the MJCF file.

    Returns:
        The parent body element, or None if not found.
    """
    # Find all body elements
    for body in root.findall(".//body"):
        # Check if the joint is a direct child of this body
        for child in body:
            if child == joint:
                return body
    return None


def add_backlash_joints(root: ET.Element) -> None:
    """Add backlash joints to all hinge joints.

    Args:
        root: The root element of the MJCF file.
    """
    # Find all hinge joints
    hinge_joints = root.findall(".//joint[@type='hinge']")

    for joint in hinge_joints:
        # Skip if this is already a backlash joint
        if joint.get("class") == "backlash":
            continue

        # Get the joint name and parent body
        joint_name = joint.get("name")
        if joint_name is None:
            continue

        parent_body = find_parent_body(joint, root)
        if parent_body is None:
            continue

        # Create a backlash joint with the same axis
        backlash_joint = ET.Element(
            "joint",
            attrib={"name": f"{joint_name}_backlash", "class": "backlash"},
        )

        # Copy the axis attribute if it exists
        if (axis := joint.get("axis")) is not None:
            backlash_joint.set("axis", axis)

        # Copy the joint position and quaternion.
        if (pos := joint.get("pos")) is not None:
            backlash_joint.set("pos", pos)
        if (quat := joint.get("quat")) is not None:
            backlash_joint.set("quat", quat)

        # Add the backlash joint after the original joint
        # Find the index of the joint in the parent body's children
        joint_index = -1
        for i, child in enumerate(parent_body):
            if child == joint:
                joint_index = i
                break

        if joint_index >= 0:
            parent_body.insert(joint_index + 1, backlash_joint)
            logger.info("Added backlash joint for %s", joint_name)


def add_backlash(mjcf_path: str | Path, backlash: float, backlash_damping: float) -> None:
    """Add backlash joints to hinge joints in the MJCF file.

    Args:
        mjcf_path: The path to the MJCF file to process.
        backlash: The amount of backlash to add.
        backlash_damping: The damping value for backlash joints.
    """
    tree = ET.parse(mjcf_path)
    root = tree.getroot()
    add_backlash_default(root, backlash, backlash_damping)
    add_backlash_joints(root)
    save_xml(mjcf_path, tree)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("mjcf_path", type=Path)
    parser.add_argument("backlash", type=float)
    parser.add_argument("--damping", type=float, default=0.01)
    args = parser.parse_args()

    add_backlash(args.mjcf_path, args.backlash, args.damping)


if __name__ == "__main__":
    # python -m urdf2mjcf.postprocess.add_backlash
    main()



"""Defines a post-processing function that handles base joints correctly."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from urdf2mjcf.utils import save_xml

logger = logging.getLogger(__name__)


def fix_base_joint(mjcf_path: str | Path, add_freejoint: bool = True) -> None:
    """Fixes the base joint configuration.

    Args:
        mjcf_path: Path to the MJCF file
        add_freejoint: Whether to add a freejoint to the root body
    """
    tree = ET.parse(mjcf_path)
    root = tree.getroot()
    worldbody = root.find("worldbody")
    if worldbody is None:
        return

    # Find the robot root body
    robot_body = worldbody.find("body")
    if robot_body is None:
        logger.warning("No robot body found in worldbody")
        return

    has_joint = robot_body.find("joint") is not None

    if has_joint:
        logger.warning("Robot body already has a joint; adding a new root body")

        # Create new root body with free joint
        new_root = ET.Element(
            "body",
            attrib={
                "name": "root",
                "pos": robot_body.get("pos", "0 0 0"),
                "quat": robot_body.get("quat", "1 0 0 0"),
            },
        )

        if add_freejoint:
            ET.SubElement(new_root, "freejoint", attrib={"name": "floating_base"})

        # Move robot body under new root
        worldbody.remove(robot_body)
        robot_body.attrib["pos"] = "0 0 0"  # Reset position relative to new root
        robot_body.attrib["quat"] = "1 0 0 0"  # Reset orientation relative to new root
        new_root.append(robot_body)
        worldbody.insert(0, new_root)

    elif add_freejoint:
        logger.warning("Robot body does not have a joint; adding a freejoint")
        robot_body.insert(0, ET.Element("freejoint", attrib={"name": "floating_base"}))

    # Removes the base inertial element.
    if (base_inertial := robot_body.find("inertial")) is not None:
        robot_body.remove(base_inertial)

    # Save changes
    save_xml(mjcf_path, tree)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("mjcf_path", type=Path)
    args = parser.parse_args()

    fix_base_joint(args.mjcf_path)


if __name__ == "__main__":
    # python -m urdf2mjcf.postprocess.base_joint
    main()



"""Defines a post-processing function that adds sensors to the MJCF model."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from scipy.spatial.transform import Rotation as R

from urdf2mjcf.model import ConversionMetadata
from urdf2mjcf.utils import save_xml

logger = logging.getLogger(__name__)


def add_sensors(
    mjcf_path: str | Path,
    root_body_name: str,
    metadata: ConversionMetadata | None = None,
) -> None:
    """Add sensors to the MJCF model.

    Args:
        mjcf_path: Path to the MJCF file
        root_body_name: Name of the root body
        metadata: Metadata for the MJCF model
    """
    if metadata is None:
        metadata = ConversionMetadata()

    tree = ET.parse(mjcf_path)
    mjcf_root = tree.getroot()

    sensor_elem = mjcf_root.find("sensor")
    if sensor_elem is None:
        sensor_elem = ET.SubElement(mjcf_root, "sensor")

    def add_sensors(site_name: str) -> None:
        ET.SubElement(
            sensor_elem,
            "framepos",
            attrib={
                "name": f"{site_name}_pos",
                "objtype": "site",
                "objname": site_name,
            },
        )
        ET.SubElement(
            sensor_elem,
            "framequat",
            attrib={
                "name": f"{site_name}_quat",
                "objtype": "site",
                "objname": site_name,
            },
        )
        ET.SubElement(
            sensor_elem,
            "framelinvel",
            attrib={
                "name": f"{site_name}_linvel",
                "objtype": "site",
                "objname": site_name,
            },
        )
        ET.SubElement(
            sensor_elem,
            "frameangvel",
            attrib={
                "name": f"{site_name}_angvel",
                "objtype": "site",
                "objname": site_name,
            },
        )
        ET.SubElement(
            sensor_elem,
            "velocimeter",
            attrib={
                "name": f"{site_name}_vel",
                "site": site_name,
            },
        )

    # Finds the root body.
    root_body = mjcf_root.find(f".//body[@name='{root_body_name}']")
    if root_body is None:
        raise ValueError(f"Root body {root_body_name} not found in the MJCF model.")

    # Find the site associated with the root body.
    site_elem = root_body.find(".//site")
    if site_elem is None:
        site_elem = ET.SubElement(root_body, "site", name=f"{root_body_name}_site")
    site_name = site_elem.attrib["name"]

    add_sensors(site_name)

    if metadata.imus:
        for imu in metadata.imus:
            # Find the link to attach the IMU to
            link_body = mjcf_root.find(f".//body[@name='{imu.body_name}']")
            if link_body is None:
                options = [body.attrib["name"] for body in mjcf_root.findall(".//body")]
                raise ValueError(f"Body {imu.body_name} not found for IMU sensor. Options: {options}")

            # Find the site associated with the link.
            site_elem = link_body.find(".//site")
            if site_elem is None:
                site_elem = ET.SubElement(link_body, "site", name=f"{imu.body_name}_site")
            site_name = site_elem.attrib["name"]

            # Updates the site position and rotation.
            if imu.rpy is not None:
                rotation = R.from_euler("xyz", imu.rpy, degrees=True)
                qx, qy, qz, qw = rotation.as_quat(scalar_first=False)
                site_elem.attrib["quat"] = f"{qw} {qx} {qy} {qz}"

            if imu.pos is not None:
                site_elem.attrib["pos"] = " ".join(str(x) for x in imu.pos)

            # Add the accelerometer
            acc_attrib = {
                "name": f"{imu.body_name}_acc",
                "site": site_name,
            }
            if imu.acc_noise is not None:
                acc_attrib["noise"] = str(imu.acc_noise)
            ET.SubElement(sensor_elem, "accelerometer", attrib=acc_attrib)

            # Add the gyroscope
            gyro_attrib = {
                "name": f"{imu.body_name}_gyro",
                "site": site_name,
            }
            if imu.gyro_noise is not None:
                gyro_attrib["noise"] = str(imu.gyro_noise)
            ET.SubElement(sensor_elem, "gyro", attrib=gyro_attrib)

            # Add the magnetometer
            mag_attrib = {
                "name": f"{imu.body_name}_mag",
                "site": site_name,
            }
            if imu.mag_noise is not None:
                mag_attrib["noise"] = str(imu.mag_noise)
            ET.SubElement(sensor_elem, "magnetometer", attrib=mag_attrib)

            # Add other sensors
            add_sensors(site_name)

    # Find the first <body> element to attach the default cameras instead of the root element.
    first_body = mjcf_root.find(".//body")
    if first_body is None:
        raise ValueError("No <body> element found in the MJCF model to attach cameras.")

    for cam in metadata.cameras:
        attrib = {
            "name": cam.name,
            "mode": cam.mode,
            "fovy": str(cam.fovy),
        }

        if cam.rpy is not None:
            rotation = R.from_euler("xyz", cam.rpy, degrees=True)
            qx, qy, qz, qw = rotation.as_quat(scalar_first=False)
            attrib["quat"] = f"{qw} {qx} {qy} {qz}"

        if cam.pos is not None:
            attrib["pos"] = " ".join(str(x) for x in cam.pos)

        ET.SubElement(first_body, "camera", attrib=attrib)

    # Add force sensors.
    for fs in metadata.force_sensors:
        # Find the body to attach the site to
        link_body = mjcf_root.find(f".//body[@name='{fs.body_name}']")
        if link_body is None:
            options = [body.attrib["name"] for body in mjcf_root.findall(".//body")]
            logger.warning(
                f"Body '{fs.body_name}' not found for force sensor site '{fs.site_name}'. "
                f"Skipping sensor. Available bodies: {options}"
            )
            continue

        # Find or create the site within the body
        site_elem = link_body.find(f"./site[@name='{fs.site_name}']")
        if site_elem is None:
            site_elem = ET.SubElement(link_body, "site", name=fs.site_name)
            logger.info(f"Created site '{fs.site_name}' on body '{fs.body_name}' for force sensor.")

        # Add the force sensor element
        fs_name = fs.name if fs.name else f"{fs.site_name}_force"
        fs_attrib = {
            "name": fs_name,
            "site": fs.site_name,  # Use the site_name directly
        }
        if fs.noise is not None:
            fs_attrib["noise"] = str(fs.noise)

        ET.SubElement(sensor_elem, "force", attrib=fs_attrib)

    # Save changes
    save_xml(mjcf_path, tree)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("mjcf_path", type=Path)
    args = parser.parse_args()

    add_sensors(args.mjcf_path, "base")


if __name__ == "__main__":
    # python -m urdf2mjcf.postprocess.add_sensors
    main()



"""Converts some mesh geometries into easier-to-collide shapes.

For frameworks like MJX, it is usually a good idea to model some of the links
as simpler shapes, since it is more robust to collide them with other links in
the scene.
"""

import argparse
import itertools
import logging
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Sequence

import numpy as np
import trimesh
from scipy.spatial.transform import Rotation as R

from urdf2mjcf.model import CollisionGeometry, CollisionType
from urdf2mjcf.utils import save_xml

logger = logging.getLogger(__name__)


def update_collisions(
    mjcf_path: str | Path,
    collision_geometries: Sequence[CollisionGeometry],
    class_name: str = "collision",
) -> None:
    """Converts some mesh geometries into easier-to-collide shapes.

    Args:
        mjcf_path: Path to the MJCF file.
        collision_geometries: List of collision geometries to process.
        class_name: The class name to use for the sphere geoms.
    """
    mjcf_path = Path(mjcf_path)
    tree = ET.parse(mjcf_path)
    root = tree.getroot()

    # Get all the meshes from the <asset> element.
    asset = root.find("asset")
    if asset is None:
        raise ValueError("No <asset> element found in the MJCF file.")
    meshes = asset.findall("mesh")
    mesh_name_to_path = {
        mesh.attrib.get("name", mesh.attrib.get("file", "MISSING")): mesh.attrib["file"] for mesh in meshes
    }

    name_to_geom = {g.name: g for g in collision_geometries}
    link_set = set(name_to_geom.keys())

    # Iterate over all <body> elements and process those in foot_links.
    for body_elem in root.iter("body"):
        body_name = body_elem.attrib.get("name", "")
        if body_name not in link_set:
            continue
        link_set.remove(body_name)
        collision_geom_info = name_to_geom[body_name]

        # Find the mesh geom in the body, disambiguating by class if necessary.
        mesh_geoms = [geom for geom in body_elem.findall("geom") if geom.attrib.get("type", "").lower() == "mesh"]
        if len(mesh_geoms) == 0:
            raise ValueError(f"No mesh geom found in link {body_name}")
        if len(mesh_geoms) > 1:
            logger.warning("Got multiple mesh geoms in link %s; attempting to use class %s", body_name, class_name)
            mesh_geoms = [geom for geom in mesh_geoms if geom.attrib.get("class", "").lower() == class_name]

            if len(mesh_geoms) == 0:
                raise ValueError(f"No mesh geom with class {class_name} found in link {body_name}")
            if len(mesh_geoms) > 1:
                raise ValueError(f"Got multiple mesh geoms with class {class_name} in link {body_name}")

        mesh_geom = mesh_geoms[0]
        mesh_geom_name = mesh_geom.attrib.get("name")

        # Find any visual meshes in this body to get material from - using naming convention
        visual_mesh_name = f"{body_name}_visual"
        visual_meshes = [geom for geom in body_elem.findall("geom") if geom.attrib.get("name") == visual_mesh_name]
        if len(visual_meshes) == 0:
            logger.warning(
                "No visual mesh found for %s in body %s."
                "Box collision will be added, but corresponding visual will not be updated.",
                visual_mesh_name,
                body_name,
            )
            visual_mesh = None
        else:
            visual_mesh = visual_meshes[0]

        mesh_name = mesh_geom.attrib.get("mesh")
        if not mesh_name:
            logger.warning("Mesh geom in link %s does not specify a mesh file; skipping.", body_name)
            continue

        if mesh_name not in mesh_name_to_path:
            logger.warning("Mesh name %s not found in <asset> element; skipping.", mesh_name)
            continue
        mesh_file = mesh_name_to_path[mesh_name]

        # Load the mesh using trimesh.
        mesh_path = (mjcf_path.parent / mesh_file).resolve()
        try:
            mesh = trimesh.load(mesh_path)
        except Exception as e:
            logger.error("Failed to load mesh from %s for link %s: %s", mesh_path, body_name, e)
            continue

        if not isinstance(mesh, trimesh.Trimesh):
            logger.warning("Loaded mesh from %s is not a Trimesh for link %s; skipping.", mesh_path, body_name)
            continue

        # Transform the mesh vertices to world coordinates.
        vertices = mesh.vertices  # shape (n,3)

        # Find geom by name in the XML and use its attributes
        geom_pos = np.zeros(3, dtype=np.float64)
        geom_quat = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)  # Default identity quaternion

        # Get position and orientation from the mesh geom XML
        if "pos" in mesh_geom.attrib:
            pos_values = [float(v) for v in mesh_geom.attrib["pos"].split()]
            geom_pos[:] = pos_values  # Update values in-place

        if "quat" in mesh_geom.attrib:
            quat_values = [float(v) for v in mesh_geom.attrib["quat"].split()]
            geom_quat[:] = quat_values  # Update values in-place

        # Get rotation matrix from quaternion
        geom_r = R.from_quat(geom_quat, scalar_first=True)

        # Transform vertices to mesh-local coordinates
        local_vertices = vertices.copy()

        # Transform vertices to account for geom's local position and orientation
        local_vertices = geom_r.apply(vertices) + geom_pos

        # Compute bounding box in local coordinates
        min_x, min_y, min_z = local_vertices.min(axis=0)
        max_x, max_y, max_z = local_vertices.max(axis=0)

        # Optional per-axis offset to translate the generated capsules.
        offset = np.array(
            [
                collision_geom_info.offset_x,
                collision_geom_info.offset_y,
                collision_geom_info.offset_z,
            ],
            dtype=np.float64,
        )

        match collision_geom_info.collision_type:
            case CollisionType.BOX:
                # Create box with same dimensions as original mesh bounding box
                box_size = np.array(
                    [
                        (max_x - min_x) / 2,
                        (max_y - min_y) / 2,
                        (max_z - min_z) / 2,
                    ]
                )

                # Position at center of bounding box
                box_pos = np.array([(max_x + min_x) / 2, (max_y + min_y) / 2, (max_z + min_z) / 2])

                # Use the original geom's orientation
                box_quat = geom_quat

                # Add a bounding box geom.
                box_geom = ET.Element("geom")
                box_geom.attrib["name"] = f"{mesh_geom_name}_box"
                box_geom.attrib["type"] = "box"
                box_geom.attrib["pos"] = " ".join(f"{v:.6f}" for v in box_pos)
                box_geom.attrib["quat"] = " ".join(f"{v:.6f}" for v in box_quat)
                box_geom.attrib["size"] = " ".join(f"{v:.6f}" for v in box_size)
                box_geom.attrib["material"] = "collision_material"

                # Copies over any other attributes from the original mesh geom.
                for key in ("class", "condim", "solref", "solimp", "fluidshape", "fluidcoef", "margin"):
                    if key in mesh_geom.attrib:
                        box_geom.attrib[key] = mesh_geom.attrib[key]

                body_elem.append(box_geom)

                # Update the visual mesh to be a box instead of creating a new one
                # Replace the mesh with a box
                if visual_mesh is not None:
                    visual_mesh.attrib["type"] = "box"
                    visual_mesh.attrib["pos"] = " ".join(f"{v:.6f}" for v in box_pos)
                    visual_mesh.attrib["quat"] = " ".join(f"{v:.6f}" for v in box_quat)
                    visual_mesh.attrib["size"] = " ".join(f"{v:.6f}" for v in box_size)

                    # Remove mesh attribute as it's now a box
                    if "mesh" in visual_mesh.attrib:
                        del visual_mesh.attrib["mesh"]

                    logger.info("Updated visual mesh %s to be a box", visual_mesh_name)

            case CollisionType.PARALLEL_CAPSULES:
                rad = collision_geom_info.sphere_radius
                ax = collision_geom_info.axis_order
                flip_axis = collision_geom_info.flip_axis

                pairs = [(min_x, max_x), (min_y, max_y), (min_z, max_z)]

                # Determine which direction is the longest
                length = [max_x - min_x, max_y - min_y, max_z - min_z][ax[0]]

                # Use the original geom's orientation
                sphere_quat = geom_quat

                for i, min_side in enumerate((True, False)):
                    min_s, max_s = pairs[ax[1]]
                    range_s = (min_s, min_s + rad * 2) if min_side else (max_s - rad * 2, max_s)

                    mask = (local_vertices[:, ax[1]] >= range_s[0]) & (local_vertices[:, ax[1]] <= range_s[1])
                    min_v = np.min(local_vertices[mask, ax[0]])
                    max_v = np.max(local_vertices[mask, ax[0]])

                    capsule_fromto = np.zeros(6, dtype=np.float64)
                    capsule_fromto[ax[0]] = min_v + rad
                    capsule_fromto[ax[0] + 3] = max_v - rad

                    val_1 = pairs[ax[1]][0] + rad if min_side else pairs[ax[1]][1] - rad
                    capsule_fromto[ax[1]] = val_1
                    capsule_fromto[ax[1] + 3] = val_1

                    val_2 = pairs[ax[2]][0] + rad if flip_axis else pairs[ax[2]][1] - rad
                    capsule_fromto[ax[2]] = val_2
                    capsule_fromto[ax[2] + 3] = val_2

                    # Apply global offset
                    capsule_fromto[:3] += offset
                    capsule_fromto[3:] += offset

                    # Create the capsule geom
                    capsule_geom = ET.Element("geom")
                    capsule_geom.attrib["name"] = f"{mesh_geom_name}_capsule_{i}"
                    capsule_geom.attrib["type"] = "capsule"
                    capsule_geom.attrib["quat"] = " ".join(f"{v:.6f}" for v in sphere_quat)
                    capsule_geom.attrib["fromto"] = " ".join(f"{v:.6f}" for v in capsule_fromto)
                    capsule_geom.attrib["size"] = f"{rad:.6f} {length / 2:.6f}"
                    capsule_geom.attrib["material"] = "collision_material"

                    # Copy over any other attributes from the original mesh geom
                    for key in ("class", "condim", "solref", "solimp", "fluidshape", "fluidcoef", "margin"):
                        if key in mesh_geom.attrib:
                            capsule_geom.attrib[key] = mesh_geom.attrib[key]

                    body_elem.append(capsule_geom)

                    # Update the visual mesh to be capsules instead of creating new ones
                    if visual_mesh is not None:
                        visual_capsule = ET.Element("geom")
                        visual_capsule.attrib["name"] = f"{visual_mesh_name}_capsule_{i}"
                        visual_capsule.attrib["type"] = "capsule"
                        visual_capsule.attrib["quat"] = " ".join(f"{v:.6f}" for v in sphere_quat)
                        visual_capsule.attrib["fromto"] = " ".join(f"{v:.6f}" for v in capsule_fromto)
                        visual_capsule.attrib["size"] = f"{rad:.6f} {length / 2:.6f}"

                        # Copy over material and class attributes
                        for key in ("material", "class"):
                            if key in visual_mesh.attrib:
                                visual_capsule.attrib[key] = visual_mesh.attrib[key]

                        body_elem.append(visual_capsule)

                if visual_mesh is not None:
                    logger.info("Updated visual mesh %s to be capsules", visual_mesh_name)

            case CollisionType.CORNER_SPHERES:
                rad = collision_geom_info.sphere_radius
                ax = collision_geom_info.axis_order
                flip_axis = collision_geom_info.flip_axis

                pairs = [(min_x, max_x), (min_y, max_y), (min_z, max_z)]

                # Use the original geom's orientation
                sphere_quat = geom_quat

                for i, (min_x, min_y) in enumerate(itertools.product((True, False), (True, False))):
                    x = pairs[ax[0]][0] + rad if min_x else pairs[ax[0]][1] - rad
                    y = pairs[ax[1]][0] + rad if min_y else pairs[ax[1]][1] - rad
                    z = pairs[ax[2]][0] + rad if flip_axis else pairs[ax[2]][1] - rad

                    xyz = [0, 0, 0]
                    xyz[ax[0]] = x
                    xyz[ax[1]] = y
                    xyz[ax[2]] = z
                    x, y, z = xyz

                    # Create the capsule geom
                    sphere_geom = ET.Element("geom")
                    sphere_geom.attrib["name"] = f"{mesh_geom_name}_sphere_{i}"
                    sphere_geom.attrib["type"] = "sphere"
                    sphere_geom.attrib["quat"] = " ".join(f"{v:.6f}" for v in sphere_quat)
                    sphere_geom.attrib["pos"] = " ".join(f"{v:.6f}" for v in (x, y, z))
                    sphere_geom.attrib["size"] = f"{rad:.6f}"
                    sphere_geom.attrib["material"] = "collision_material"

                    # Copy over any other attributes from the original mesh geom
                    for key in ("class", "condim", "solref", "solimp", "fluidshape", "fluidcoef", "margin"):
                        if key in mesh_geom.attrib:
                            sphere_geom.attrib[key] = mesh_geom.attrib[key]

                    body_elem.append(sphere_geom)

                    # Update the visual mesh to be capsules instead of creating new ones
                    if visual_mesh is not None:
                        visual_sphere = ET.Element("geom")
                        visual_sphere.attrib["name"] = f"{visual_mesh_name}_sphere_{i}"
                        visual_sphere.attrib["type"] = "sphere"
                        visual_sphere.attrib["quat"] = " ".join(f"{v:.6f}" for v in sphere_quat)
                        visual_sphere.attrib["pos"] = " ".join(f"{v:.6f}" for v in (x, y, z))
                        visual_sphere.attrib["size"] = f"{rad:.6f}"

                        # Copy over material and class attributes
                        for key in ("material", "class"):
                            if key in visual_mesh.attrib:
                                visual_sphere.attrib[key] = visual_mesh.attrib[key]

                        body_elem.append(visual_sphere)

                if visual_mesh is not None:
                    body_elem.remove(visual_mesh)
                    logger.info("Updated visual mesh %s to be corner spheres", visual_mesh_name)

            case CollisionType.SINGLE_SPHERE:
                rad = collision_geom_info.sphere_radius
                ax = collision_geom_info.axis_order
                flip_axis = collision_geom_info.flip_axis
                pairs = [(min_x, max_x), (min_y, max_y), (min_z, max_z)]

                min_v, max_v = pairs[ax[2]]
                pairs[ax[2]] = (min_v + rad, min_v + rad) if flip_axis else (max_v - rad, max_v - rad)

                sphere_pos = np.array([sum(pairs[0]) / 2, sum(pairs[1]) / 2, sum(pairs[2]) / 2])
                sphere_quat = geom_quat

                sphere_geom = ET.Element("geom")
                sphere_geom.attrib["name"] = f"{mesh_geom_name}_sphere"
                sphere_geom.attrib["type"] = "sphere"
                sphere_geom.attrib["quat"] = " ".join(f"{v:.6f}" for v in sphere_quat)
                sphere_geom.attrib["pos"] = " ".join(f"{v:.6f}" for v in sphere_pos)
                sphere_geom.attrib["size"] = f"{rad:.6f}"
                sphere_geom.attrib["material"] = "collision_material"

                body_elem.append(sphere_geom)

                if visual_mesh is not None:
                    visual_mesh.attrib["type"] = "sphere"
                    visual_mesh.attrib["pos"] = " ".join(f"{v:.6f}" for v in sphere_pos)
                    visual_mesh.attrib["quat"] = " ".join(f"{v:.6f}" for v in sphere_quat)
                    visual_mesh.attrib["size"] = f"{rad:.6f}"

                    # Remove mesh attribute as it's now a sphere
                    if "mesh" in visual_mesh.attrib:
                        del visual_mesh.attrib["mesh"]

                    logger.info("Updated visual mesh %s to be a single sphere", visual_mesh_name)

            case _:
                raise NotImplementedError(f"Collision type {collision_geom_info.collision_type} not implemented.")

        # Remove the original mesh geom from the body.
        body_elem.remove(mesh_geom)

    if link_set:
        raise ValueError(f"Found {len(link_set)} collision geometries that were not found in the MJCF file: {link_set}")

    # Save the modified MJCF file.
    save_xml(mjcf_path, tree)
    logger.info("Saved modified MJCF file with feet converted to boxes at %s", mjcf_path)


def main() -> None:
    parser = argparse.ArgumentParser(description="Converts MJCF collision geometries from meshes to boxes.")
    parser.add_argument("mjcf_path", type=Path, help="Path to the MJCF file.")
    parser.add_argument("--links", nargs="+", required=True, help="List of link names to convert.")
    args = parser.parse_args()

    collision_geometries = [CollisionGeometry(name=name, collision_type=CollisionType.BOX) for name in args.links]
    update_collisions(args.mjcf_path, collision_geometries)


if __name__ == "__main__":
    main()



"""Defines a dummy test."""

import json
import tempfile
from pathlib import Path

import pytest

from urdf2mjcf.convert import convert_urdf_to_mjcf
from urdf2mjcf.model import ActuatorMetadata, JointMetadata


@pytest.mark.slow
def test_conversion_output(tmpdir: Path) -> None:
    urdf_path = Path(__file__).parent / "sample" / "robot.urdf"
    mjcf_path = tmpdir / "robot.mjcf"

    # Load joint metadata
    joint_metadata_path = urdf_path.parent / "joint_metadata.json"
    with open(joint_metadata_path, "r") as f:
        joint_metadata = json.load(f)["joint_name_to_metadata"]
        for key, value in joint_metadata.items():
            joint_metadata[key] = JointMetadata.from_dict(value)

    # Load actuator metadata
    actuator_path = urdf_path.parent / "actuators" / "motor.json"
    with open(actuator_path, "r") as f:
        motor_data = json.load(f)
        actuator_type = motor_data["actuator_type"]
        actuator_metadata = {actuator_type: ActuatorMetadata.from_dict(motor_data)}

    convert_urdf_to_mjcf(
        urdf_path=urdf_path,
        mjcf_path=mjcf_path,
        copy_meshes=False,
        metadata_file=urdf_path.parent / "metadata.json",
        joint_metadata=joint_metadata,
        actuator_metadata=actuator_metadata,
    )

    # After making a change, put a breakpoint here and make sure you try out
    # the model in Mujoco before committing changes.
    assert mjcf_path.exists()


if __name__ == "__main__":
    # python -m tests.test_conversion
    with tempfile.TemporaryDirectory() as temp_dir:
        test_conversion_output(Path(temp_dir))
